@@ -1,7 +1,8 @@
 /* Capstone Disassembly Engine */
-/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2014 */
+/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2015 */
 
 #if defined(CAPSTONE_HAS_OSXKERNEL)
+#include <Availability.h>
 #include <libkern/libkern.h>
 #else
 #include <stdio.h>
@@ -16,11 +17,21 @@
 
 void MCInst_Init(MCInst *inst)
 {
+	unsigned int i;
+
+	for (i = 0; i < 48; i++) {
+		inst->Operands[i].Kind = kInvalid;
+	}
+
+	inst->Opcode = 0;
 	inst->OpcodePub = 0;
 	inst->size = 0;
 	inst->has_imm = false;
 	inst->op1_size = 0;
 	inst->writeback = false;
+	inst->ac_idx = 0;
+	inst->popcode_adjust = 0;
+	inst->assembly[0] = '\0';
 }
 
 void MCInst_clear(MCInst *inst)
