@@ -13,15 +13,22 @@
 //===----------------------------------------------------------------------===//
 
 /* Capstone Disassembly Engine */
-/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2014 */
+/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2015 */
 
 #ifdef CAPSTONE_HAS_X86
 
+#if defined (WIN32) || defined (WIN64) || defined (_WIN32) || defined (_WIN64)
+#pragma warning(disable:4996)			// disable MSVC's warning on strncpy()
+#pragma warning(disable:28719)		// disable MSVC's warning on strncpy()
+#endif
+
 #if !defined(CAPSTONE_HAS_OSXKERNEL)
 #include <ctype.h>
 #endif
-#include "../../myinttypes.h"
+#include <capstone/platform.h>
+
 #if defined(CAPSTONE_HAS_OSXKERNEL)
+#include <Availability.h>
 #include <libkern/libkern.h>
 #else
 #include <stdio.h>
@@ -63,17 +70,71 @@ static void set_mem_access(MCInst *MI, bool status)
 
 static void printopaquemem(MCInst *MI, unsigned OpNo, SStream *O)
 {
-	SStream_concat0(O, "ptr ");
+	// FIXME: do this with autogen
+	// printf(">>> ID = %u\n", MI->flat_insn->id);
+	switch(MI->flat_insn->id) {
+		default:
+			SStream_concat0(O, "ptr ");
+			break;
+		case X86_INS_SGDT:
+		case X86_INS_SIDT:
+		case X86_INS_LGDT:
+		case X86_INS_LIDT:
+		case X86_INS_FXRSTOR:
+		case X86_INS_FXSAVE:
+		case X86_INS_LJMP:
+		case X86_INS_LCALL:
+			// do not print "ptr"
+			break;
+	}
 
 	switch(MI->csh->mode) {
 		case CS_MODE_16:
-			MI->x86opsize = 2;
+			switch(MI->flat_insn->id) {
+				default:
+					MI->x86opsize = 2;
+					break;
+				case X86_INS_LJMP:
+				case X86_INS_LCALL:
+					MI->x86opsize = 4;
+					break;
+				case X86_INS_SGDT:
+				case X86_INS_SIDT:
+				case X86_INS_LGDT:
+				case X86_INS_LIDT:
+					MI->x86opsize = 6;
+					break;
+			}
 			break;
 		case CS_MODE_32:
-			MI->x86opsize = 4;
+			switch(MI->flat_insn->id) {
+				default:
+					MI->x86opsize = 4;
+					break;
+				case X86_INS_LJMP:
+				case X86_INS_LCALL:
+				case X86_INS_SGDT:
+				case X86_INS_SIDT:
+				case X86_INS_LGDT:
+				case X86_INS_LIDT:
+					MI->x86opsize = 6;
+					break;
+			}
 			break;
 		case CS_MODE_64:
-			MI->x86opsize = 8;
+			switch(MI->flat_insn->id) {
+				default:
+					MI->x86opsize = 8;
+					break;
+				case X86_INS_LJMP:
+				case X86_INS_LCALL:
+				case X86_INS_SGDT:
+				case X86_INS_SIDT:
+				case X86_INS_LGDT:
+				case X86_INS_LIDT:
+					MI->x86opsize = 10;
+					break;
+			}
 			break;
 		default:	// never reach
 			break;
@@ -134,8 +195,34 @@ static void printi512mem(MCInst *MI, unsigned OpNo, SStream *O)
 
 static void printf32mem(MCInst *MI, unsigned OpNo, SStream *O)
 {
-	SStream_concat0(O, "dword ptr ");
-	MI->x86opsize = 4;
+	switch(MCInst_getOpcode(MI)) {
+		default:
+			SStream_concat0(O, "dword ptr ");
+			MI->x86opsize = 4;
+			break;
+		case X86_FBSTPm:
+		case X86_FBLDm:
+			// TODO: fix this in tablegen instead
+			SStream_concat0(O, "tbyte ptr ");
+			MI->x86opsize = 10;
+			break;
+		case X86_FSTENVm:
+		case X86_FLDENVm:
+			// TODO: fix this in tablegen instead
+			switch(MI->csh->mode) {
+				default:    // never reach
+					break;
+				case CS_MODE_16:
+					MI->x86opsize = 14;
+					break;
+				case CS_MODE_32:
+				case CS_MODE_64:
+					MI->x86opsize = 28;
+					break;
+			}
+			break;
+	}
+
 	printMemReference(MI, OpNo, O);
 }
 
@@ -176,7 +263,7 @@ static void printf512mem(MCInst *MI, unsigned OpNo, SStream *O)
 
 static void printSSECC(MCInst *MI, unsigned Op, SStream *OS)
 {
-	int64_t Imm = MCOperand_getImm(MCInst_getOperand(MI, Op)) & 7;
+	uint8_t Imm = (uint8_t)(MCOperand_getImm(MCInst_getOperand(MI, Op)) & 7);
 	switch (Imm) {
 		default: break;	// never reach
 		case    0: SStream_concat0(OS, "eq"); op_addSseCC(MI, X86_SSE_CC_EQ); break;
@@ -187,20 +274,14 @@ static void printSSECC(MCInst *MI, unsigned Op, SStream *OS)
 		case    5: SStream_concat0(OS, "nlt"); op_addSseCC(MI, X86_SSE_CC_NLT); break;
 		case    6: SStream_concat0(OS, "nle"); op_addSseCC(MI, X86_SSE_CC_NLE); break;
 		case    7: SStream_concat0(OS, "ord"); op_addSseCC(MI, X86_SSE_CC_ORD); break;
-		case    8: SStream_concat0(OS, "eq_uq"); op_addSseCC(MI, X86_SSE_CC_EQ_UQ); break;
-		case    9: SStream_concat0(OS, "nge"); op_addSseCC(MI, X86_SSE_CC_NGE); break;
-		case  0xa: SStream_concat0(OS, "ngt"); op_addSseCC(MI, X86_SSE_CC_NGT); break;
-		case  0xb: SStream_concat0(OS, "false"); op_addSseCC(MI, X86_SSE_CC_FALSE); break;
-		case  0xc: SStream_concat0(OS, "neq_oq"); op_addSseCC(MI, X86_SSE_CC_NEQ_OQ); break;
-		case  0xd: SStream_concat0(OS, "ge"); op_addSseCC(MI, X86_SSE_CC_GE); break;
-		case  0xe: SStream_concat0(OS, "gt"); op_addSseCC(MI, X86_SSE_CC_GT); break;
-		case  0xf: SStream_concat0(OS, "true"); op_addSseCC(MI, X86_SSE_CC_TRUE); break;
 	}
+
+	MI->popcode_adjust = Imm + 1;
 }
 
 static void printAVXCC(MCInst *MI, unsigned Op, SStream *O)
 {
-	int64_t Imm = MCOperand_getImm(MCInst_getOperand(MI, Op)) & 0x1f;
+	uint8_t Imm = (uint8_t)(MCOperand_getImm(MCInst_getOperand(MI, Op)) & 0x1f);
 	switch (Imm) {
 		default: break;//printf("Invalid avxcc argument!\n"); break;
 		case    0: SStream_concat0(O, "eq"); op_addAvxCC(MI, X86_AVX_CC_EQ); break;
@@ -236,6 +317,25 @@ static void printAVXCC(MCInst *MI, unsigned Op, SStream *O)
 		case 0x1e: SStream_concat0(O, "gt_oq"); op_addAvxCC(MI, X86_AVX_CC_GT_OQ); break;
 		case 0x1f: SStream_concat0(O, "true_us"); op_addAvxCC(MI, X86_AVX_CC_TRUE_US); break;
 	}
+
+	MI->popcode_adjust = Imm + 1;
+}
+
+static void printXOPCC(MCInst *MI, unsigned Op, SStream *O)
+{
+	int64_t Imm = MCOperand_getImm(MCInst_getOperand(MI, Op));
+
+	switch (Imm) {
+		default: // llvm_unreachable("Invalid xopcc argument!");
+		case 0: SStream_concat0(O, "lt"); op_addXopCC(MI, X86_XOP_CC_LT); break;
+		case 1: SStream_concat0(O, "le"); op_addXopCC(MI, X86_XOP_CC_LE); break;
+		case 2: SStream_concat0(O, "gt"); op_addXopCC(MI, X86_XOP_CC_GT); break;
+		case 3: SStream_concat0(O, "ge"); op_addXopCC(MI, X86_XOP_CC_GE); break;
+		case 4: SStream_concat0(O, "eq"); op_addXopCC(MI, X86_XOP_CC_EQ); break;
+		case 5: SStream_concat0(O, "neq"); op_addXopCC(MI, X86_XOP_CC_NEQ); break;
+		case 6: SStream_concat0(O, "false"); op_addXopCC(MI, X86_XOP_CC_FALSE); break;
+		case 7: SStream_concat0(O, "true"); op_addXopCC(MI, X86_XOP_CC_TRUE); break;
+	}
 }
 
 static void printRoundingControl(MCInst *MI, unsigned Op, SStream *O)
@@ -252,12 +352,97 @@ static void printRoundingControl(MCInst *MI, unsigned Op, SStream *O)
 
 #endif
 
-static char *getRegisterName(unsigned RegNo);
+static const char *getRegisterName(unsigned RegNo);
 static void printRegName(SStream *OS, unsigned RegNo)
 {
 	SStream_concat0(OS, getRegisterName(RegNo));
 }
 
+// for MASM syntax, 0x123 = 123h, 0xA123 = 0A123h
+// this function tell us if we need to have prefix 0 in front of a number
+static bool need_zero_prefix(uint64_t imm)
+{
+	// find the first hex letter representing imm
+	while(imm >= 0x10)
+		imm >>= 4;
+
+	if (imm < 0xa)
+		return false;
+	else	// this need 0 prefix
+		return true;
+}
+
+static void printImm(int syntax, SStream *O, int64_t imm, bool positive)
+{
+	if (positive) {
+		// always print this number in positive form
+		if (syntax == CS_OPT_SYNTAX_MASM) {
+			if (imm < 0) {
+				if (imm == 0x8000000000000000LL)  // imm == -imm
+					SStream_concat0(O, "8000000000000000h");
+				else if (need_zero_prefix(imm))
+					SStream_concat(O, "0%"PRIx64"h", imm);
+				else
+					SStream_concat(O, "%"PRIx64"h", imm);
+			} else {
+				if (imm > HEX_THRESHOLD) {
+					if (need_zero_prefix(imm))
+						SStream_concat(O, "0%"PRIx64"h", imm);
+					else
+						SStream_concat(O, "%"PRIx64"h", imm);
+				} else
+					SStream_concat(O, "%"PRIu64, imm);
+			}
+		} else {	// Intel syntax
+			if (imm < 0) {
+				SStream_concat(O, "0x%"PRIx64, imm);
+			} else {
+				if (imm > HEX_THRESHOLD)
+					SStream_concat(O, "0x%"PRIx64, imm);
+				else
+					SStream_concat(O, "%"PRIu64, imm);
+			}
+		}
+	} else {
+		if (syntax == CS_OPT_SYNTAX_MASM) {
+			if (imm < 0) {
+				if (imm == 0x8000000000000000LL)  // imm == -imm
+					SStream_concat0(O, "8000000000000000h");
+				else if (imm < -HEX_THRESHOLD) {
+					if (need_zero_prefix(imm))
+						SStream_concat(O, "-0%"PRIx64"h", -imm);
+					else
+						SStream_concat(O, "-%"PRIx64"h", -imm);
+				} else
+					SStream_concat(O, "-%"PRIu64, -imm);
+			} else {
+				if (imm > HEX_THRESHOLD) {
+					if (need_zero_prefix(imm))
+						SStream_concat(O, "0%"PRIx64"h", imm);
+					else
+						SStream_concat(O, "%"PRIx64"h", imm);
+				} else
+					SStream_concat(O, "%"PRIu64, imm);
+			}
+		} else {	// Intel syntax
+			if (imm < 0) {
+				if (imm == 0x8000000000000000LL)  // imm == -imm
+					SStream_concat0(O, "0x8000000000000000");
+				else if (imm < -HEX_THRESHOLD)
+					SStream_concat(O, "-0x%"PRIx64, -imm);
+				else
+					SStream_concat(O, "-%"PRIu64, -imm);
+
+			} else {
+				if (imm > HEX_THRESHOLD)
+					SStream_concat(O, "0x%"PRIx64, imm);
+				else
+					SStream_concat(O, "%"PRIu64, imm);
+			}
+		}
+	}
+}
+
 // local printOperand, without updating public operands
 static void _printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 {
@@ -266,27 +451,47 @@ static void _printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 		printRegName(O, MCOperand_getReg(Op));
 	} else if (MCOperand_isImm(Op)) {
 		int64_t imm = MCOperand_getImm(Op);
-		if (imm < 0) {
-			if (imm < -HEX_THRESHOLD)
-				SStream_concat(O, "-0x%"PRIx64, -imm);
-			else
-				SStream_concat(O, "-%"PRIu64, -imm);
-
-		} else {
-			if (imm > HEX_THRESHOLD)
-				SStream_concat(O, "0x%"PRIx64, imm);
-			else
-				SStream_concat(O, "%"PRIu64, imm);
-		}
+		printImm(MI->csh->syntax, O, imm, MI->csh->imm_unsigned);
 	}
 }
 
+#ifndef CAPSTONE_DIET
+// copy & normalize access info
+static void get_op_access(cs_struct *h, unsigned int id, uint8_t *access, uint64_t *eflags)
+{
+#ifndef CAPSTONE_DIET
+	uint8_t i;
+	uint8_t *arr = X86_get_op_access(h, id, eflags);
+
+	if (!arr) {
+		access[0] = 0;
+		return;
+	}
+
+	// copy to access but zero out CS_AC_IGNORE
+	for(i = 0; arr[i]; i++) {
+		if (arr[i] != CS_AC_IGNORE)
+			access[i] = arr[i];
+		else
+			access[i] = 0;
+	}
+
+	// mark the end of array
+	access[i] = 0;
+#endif
+}
+#endif
+
 static void printSrcIdx(MCInst *MI, unsigned Op, SStream *O)
 {
 	MCOperand *SegReg;
 	int reg;
 
 	if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+		uint8_t access[6];
+#endif
+
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_MEM;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->x86opsize;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.segment = X86_REG_INVALID;
@@ -294,6 +499,11 @@ static void printSrcIdx(MCInst *MI, unsigned Op, SStream *O)
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.index = X86_REG_INVALID;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.scale = 1;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.disp = 0;
+
+#ifndef CAPSTONE_DIET
+		get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
 	}
 
 	SegReg = MCInst_getOperand(MI, Op+1);
@@ -318,6 +528,10 @@ static void printSrcIdx(MCInst *MI, unsigned Op, SStream *O)
 static void printDstIdx(MCInst *MI, unsigned Op, SStream *O)
 {
 	if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+		uint8_t access[6];
+#endif
+
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_MEM;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->x86opsize;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.segment = X86_REG_INVALID;
@@ -325,6 +539,11 @@ static void printDstIdx(MCInst *MI, unsigned Op, SStream *O)
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.index = X86_REG_INVALID;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.scale = 1;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.disp = 0;
+
+#ifndef CAPSTONE_DIET
+		get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
 	}
 
 	// DI accesses are always ES-based on non-64bit mode
@@ -405,6 +624,10 @@ static void printMemOffset(MCInst *MI, unsigned Op, SStream *O)
 	int reg;
 
 	if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+		uint8_t access[6];
+#endif
+
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_MEM;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->x86opsize;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.segment = X86_REG_INVALID;
@@ -412,6 +635,11 @@ static void printMemOffset(MCInst *MI, unsigned Op, SStream *O)
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.index = X86_REG_INVALID;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.scale = 1;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.disp = 0;
+
+#ifndef CAPSTONE_DIET
+		get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
 	}
 
 	// If this has a segment register, print it.
@@ -430,14 +658,11 @@ static void printMemOffset(MCInst *MI, unsigned Op, SStream *O)
 		int64_t imm = MCOperand_getImm(DispSpec);
 		if (MI->csh->detail)
 			MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.disp = imm;
-		if (imm < 0) {
-			SStream_concat(O, "0x%"PRIx64, arch_masks[MI->csh->mode] & imm);
-		} else {
-			if (imm > HEX_THRESHOLD)
-				SStream_concat(O, "0x%"PRIx64, imm);
-			else
-				SStream_concat(O, "%"PRIu64, imm);
-		}
+
+		if (imm < 0)
+			printImm(MI->csh->syntax, O, arch_masks[MI->csh->mode] & imm, true);
+		else
+			printImm(MI->csh->syntax, O, imm, true);
 	}
 
 	SStream_concat0(O, "]");
@@ -449,6 +674,31 @@ static void printMemOffset(MCInst *MI, unsigned Op, SStream *O)
 		MI->op1_size = MI->x86opsize;
 }
 
+#ifndef CAPSTONE_X86_REDUCE
+static void printU8Imm(MCInst *MI, unsigned Op, SStream *O)
+{
+	uint8_t val = MCOperand_getImm(MCInst_getOperand(MI, Op)) & 0xff;
+
+	printImm(MI->csh->syntax, O, val, true);
+
+	if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+		uint8_t access[6];
+#endif
+
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_IMM;
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].imm = val;
+
+#ifndef CAPSTONE_DIET
+		get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
+
+		MI->flat_insn->detail->x86.op_count++;
+	}
+}
+#endif
+
 static void printMemOffs8(MCInst *MI, unsigned OpNo, SStream *O)
 {
 	SStream_concat0(O, "byte ptr ");
@@ -477,22 +727,40 @@ static void printMemOffs64(MCInst *MI, unsigned OpNo, SStream *O)
 	printMemOffset(MI, OpNo, O);
 }
 
+#ifndef CAPSTONE_DIET
 static char *printAliasInstr(MCInst *MI, SStream *OS, void *info);
+#endif
 static void printInstruction(MCInst *MI, SStream *O, MCRegisterInfo *MRI);
+
 void X86_Intel_printInst(MCInst *MI, SStream *O, void *Info)
 {
+#ifndef CAPSTONE_DIET
 	char *mnem;
+#endif
 	x86_reg reg, reg2;
+	enum cs_ac_type access1, access2;
 
+	// perhaps this instruction does not need printer
+	if (MI->assembly[0]) {
+		strncpy(O->buffer, MI->assembly, sizeof(O->buffer));
+		return;
+	}
+
+#ifndef CAPSTONE_DIET
 	// Try to print any aliases first.
 	mnem = printAliasInstr(MI, O, Info);
 	if (mnem)
 		cs_mem_free(mnem);
 	else
+#endif
 		printInstruction(MI, O, Info);
 
-	reg = X86_insn_reg_intel(MCInst_getOpcode(MI));
+	reg = X86_insn_reg_intel(MCInst_getOpcode(MI), &access1);
 	if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+		uint8_t access[6] = {0};
+#endif
+
 		// first op can be embedded in the asm by llvm.
 		// so we have to add the missing register as the first operand
 		if (reg) {
@@ -502,19 +770,27 @@ void X86_Intel_printInst(MCInst *MI, SStream *O, void *Info)
 			MI->flat_insn->detail->x86.operands[0].type = X86_OP_REG;
 			MI->flat_insn->detail->x86.operands[0].reg = reg;
 			MI->flat_insn->detail->x86.operands[0].size = MI->csh->regsize_map[reg];
-			MI->flat_insn->detail->x86.operands[1].size = MI->csh->regsize_map[reg];
+			MI->flat_insn->detail->x86.operands[0].access = access1;
 			MI->flat_insn->detail->x86.op_count++;
 		} else {
-			if (X86_insn_reg_intel2(MCInst_getOpcode(MI), &reg, &reg2)) {
+			if (X86_insn_reg_intel2(MCInst_getOpcode(MI), &reg, &access1, &reg2, &access2)) {
 				MI->flat_insn->detail->x86.operands[0].type = X86_OP_REG;
 				MI->flat_insn->detail->x86.operands[0].reg = reg;
 				MI->flat_insn->detail->x86.operands[0].size = MI->csh->regsize_map[reg];
+				MI->flat_insn->detail->x86.operands[0].access = access1;
 				MI->flat_insn->detail->x86.operands[1].type = X86_OP_REG;
 				MI->flat_insn->detail->x86.operands[1].reg = reg2;
 				MI->flat_insn->detail->x86.operands[1].size = MI->csh->regsize_map[reg2];
+				MI->flat_insn->detail->x86.operands[1].access = access2;
 				MI->flat_insn->detail->x86.op_count = 2;
 			}
 		}
+
+#ifndef CAPSTONE_DIET
+		get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+		MI->flat_insn->detail->x86.operands[0].access = access[0];
+		MI->flat_insn->detail->x86.operands[1].access = access[1];
+#endif
 	}
 
 	if (MI->op1_size == 0 && reg)
@@ -528,32 +804,48 @@ static void printPCRelImm(MCInst *MI, unsigned OpNo, SStream *O)
 	MCOperand *Op = MCInst_getOperand(MI, OpNo);
 	if (MCOperand_isImm(Op)) {
 		int64_t imm = MCOperand_getImm(Op) + MI->flat_insn->size + MI->address;
+		uint8_t opsize = X86_immediate_size(MI->Opcode, NULL);
 
 		// truncat imm for non-64bit
 		if (MI->csh->mode != CS_MODE_64) {
 			imm = imm & 0xffffffff;
 		}
 
-		if (imm < 0) {
-			SStream_concat(O, "0x%"PRIx64, imm);
-		} else {
-			// handle 16bit segment bound
-			if (MI->csh->mode == CS_MODE_16 && imm > 0x100000)
-				imm -= 0x10000;
+		if (MI->csh->mode == CS_MODE_16 &&
+				(MI->Opcode != X86_JMP_4 && MI->Opcode != X86_CALLpcrel32))
+			imm = imm & 0xffff;
+
+		// Hack: X86 16bit with opcode X86_JMP_4
+		if (MI->csh->mode == CS_MODE_16 &&
+				(MI->Opcode == X86_JMP_4 && MI->x86_prefix[2] != 0x66))
+			imm = imm & 0xffff;
+
+		// CALL/JMP rel16 is special
+		if (MI->Opcode == X86_CALLpcrel16 || MI->Opcode == X86_JMP_2)
+			imm = imm & 0xffff;
+
+		printImm(MI->csh->syntax, O, imm, true);
 
-			if (imm > HEX_THRESHOLD)
-				SStream_concat(O, "0x%"PRIx64, imm);
-			else
-				SStream_concat(O, "%"PRIu64, imm);
-		}
 		if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+			uint8_t access[6];
+#endif
+
 			MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_IMM;
 			// if op_count > 0, then this operand's size is taken from the destination op
 			if (MI->flat_insn->detail->x86.op_count > 0)
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->flat_insn->detail->x86.operands[0].size;
+			else if (opsize > 0)
+				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = opsize;
 			else
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->imm_size;
 			MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].imm = imm;
+
+#ifndef CAPSTONE_DIET
+			get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+			MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
+
 			MI->flat_insn->detail->x86.op_count++;
 		}
 
@@ -564,7 +856,6 @@ static void printPCRelImm(MCInst *MI, unsigned OpNo, SStream *O)
 
 static void printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 {
-	int opsize = 0;
 	MCOperand *Op  = MCInst_getOperand(MI, OpNo);
 
 	if (MCOperand_isReg(Op)) {
@@ -575,9 +866,19 @@ static void printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 			if (MI->csh->doing_mem) {
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.base = reg;
 			} else {
+#ifndef CAPSTONE_DIET
+				uint8_t access[6];
+#endif
+
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_REG;
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].reg = reg;
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->csh->regsize_map[reg];
+
+#ifndef CAPSTONE_DIET
+				get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
+
 				MI->flat_insn->detail->x86.op_count++;
 			}
 		}
@@ -585,98 +886,40 @@ static void printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 		if (MI->op1_size == 0)
 			MI->op1_size = MI->csh->regsize_map[reg];
 	} else if (MCOperand_isImm(Op)) {
+		uint8_t encsize;
 		int64_t imm = MCOperand_getImm(Op);
+		uint8_t opsize = X86_immediate_size(MCInst_getOpcode(MI), &encsize);
 
-		switch(MCInst_getOpcode(MI)) {
-			default:
-				break;
-
-			case X86_AAD8i8:
-			case X86_AAM8i8:
-			case X86_ADC8i8:
-			case X86_ADD8i8:
-			case X86_AND8i8:
-			case X86_CMP8i8:
-			case X86_OR8i8:
-			case X86_SBB8i8:
-			case X86_SUB8i8:
-			case X86_TEST8i8:
-			case X86_XOR8i8:
-			case X86_ROL8ri:
-			case X86_ADC8ri:
-			case X86_ADD8ri:
-			case X86_ADD8ri8:
-			case X86_AND8ri:
-			case X86_AND8ri8:
-			case X86_CMP8ri:
-			case X86_MOV8ri:
-			case X86_MOV8ri_alt:
-			case X86_OR8ri:
-			case X86_OR8ri8:
-			case X86_RCL8ri:
-			case X86_RCR8ri:
-			case X86_ROR8ri:
-			case X86_SAL8ri:
-			case X86_SAR8ri:
-			case X86_SBB8ri:
-			case X86_SHL8ri:
-			case X86_SHR8ri:
-			case X86_SUB8ri:
-			case X86_SUB8ri8:
-			case X86_TEST8ri:
-			case X86_TEST8ri_NOREX:
-			case X86_TEST8ri_alt:
-			case X86_XOR8ri:
-			case X86_XOR8ri8:
-			case X86_OUT8ir:
-
-			case X86_ADC8mi:
-			case X86_ADD8mi:
-			case X86_AND8mi:
-			case X86_CMP8mi:
-			case X86_LOCK_ADD8mi:
-			case X86_LOCK_AND8mi:
-			case X86_LOCK_OR8mi:
-			case X86_LOCK_SUB8mi:
-			case X86_LOCK_XOR8mi:
-			case X86_MOV8mi:
-			case X86_OR8mi:
-			case X86_RCL8mi:
-			case X86_RCR8mi:
-			case X86_ROL8mi:
-			case X86_ROR8mi:
-			case X86_SAL8mi:
-			case X86_SAR8mi:
-			case X86_SBB8mi:
-			case X86_SHL8mi:
-			case X86_SHR8mi:
-			case X86_SUB8mi:
-			case X86_TEST8mi:
-			case X86_TEST8mi_alt:
-			case X86_XOR8mi:
-			case X86_PUSH64i8:
-			case X86_CMP32ri8:
-			case X86_CMP64ri8:
-
-				imm = imm & 0xff;
-				opsize = 1;     // immediate of 1 byte
-				break;
-		}
+		if (opsize == 1)    // print 1 byte immediate in positive form
+			imm = imm & 0xff;
 
+		// printf(">>> id = %u\n", MI->flat_insn->id);
 		switch(MI->flat_insn->id) {
 			default:
-				if (imm >= 0) {
-					if (imm > HEX_THRESHOLD)
-						SStream_concat(O, "0x%"PRIx64, imm);
-					else
-						SStream_concat(O, "%"PRIu64, imm);
-				} else {
-					if (imm < -HEX_THRESHOLD)
-						SStream_concat(O, "-0x%"PRIx64, -imm);
-					else
-						SStream_concat(O, "-%"PRIu64, -imm);
-				}
+				printImm(MI->csh->syntax, O, imm, MI->csh->imm_unsigned);
+				break;
 
+			case X86_INS_MOVABS:
+				// do not print number in negative form
+				printImm(MI->csh->syntax, O, imm, true);
+				break;
+
+			case X86_INS_IN:
+			case X86_INS_OUT:
+			case X86_INS_INT:
+				// do not print number in negative form
+				imm = imm & 0xff;
+				printImm(MI->csh->syntax, O, imm, true);
+				break;
+
+			case X86_INS_LCALL:
+			case X86_INS_LJMP:
+				// always print address in positive form
+				if (OpNo == 1) {	// ptr16 part
+					imm = imm & 0xffff;
+					opsize = 2;
+				}
+				printImm(MI->csh->syntax, O, imm, true);
 				break;
 
 			case X86_INS_AND:
@@ -684,19 +927,21 @@ static void printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 			case X86_INS_XOR:
 				// do not print number in negative form
 				if (imm >= 0 && imm <= HEX_THRESHOLD)
-					SStream_concat(O, "%u", imm);
+					printImm(MI->csh->syntax, O, imm, true);
 				else {
-					imm = arch_masks[MI->op1_size? MI->op1_size : MI->imm_size] & imm;
-					SStream_concat(O, "0x%"PRIx64, imm);
+					imm = arch_masks[opsize? opsize : MI->imm_size] & imm;
+					printImm(MI->csh->syntax, O, imm, true);
 				}
 				break;
+
 			case X86_INS_RET:
+			case X86_INS_RETF:
 				// RET imm16
 				if (imm >= 0 && imm <= HEX_THRESHOLD)
-					SStream_concat(O, "%u", imm);
+					printImm(MI->csh->syntax, O, imm, true);
 				else {
 					imm = 0xffff & imm;
-					SStream_concat(O, "0x%x", 0xffff & imm);
+					printImm(MI->csh->syntax, O, imm, true);
 				}
 				break;
 		}
@@ -705,20 +950,32 @@ static void printOperand(MCInst *MI, unsigned OpNo, SStream *O)
 			if (MI->csh->doing_mem) {
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.disp = imm;
 			} else {
+#ifndef CAPSTONE_DIET
+				uint8_t access[6];
+#endif
+
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_IMM;
-				if (opsize > 0)
+				if (opsize > 0) {
 					MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = opsize;
-				else if (MI->flat_insn->detail->x86.op_count > 0)
-					MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->flat_insn->detail->x86.operands[0].size;
-				else
+					MI->flat_insn->detail->x86.encoding.imm_size = encsize;
+				} else if (MI->flat_insn->detail->x86.op_count > 0) {
+					if (MI->flat_insn->id != X86_INS_LCALL && MI->flat_insn->id != X86_INS_LJMP) {
+						MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size =
+							MI->flat_insn->detail->x86.operands[0].size;
+					} else
+						MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->imm_size;
+				} else
 					MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->imm_size;
 				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].imm = imm;
+
+#ifndef CAPSTONE_DIET
+				get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+				MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
+
 				MI->flat_insn->detail->x86.op_count++;
 			}
 		}
-
-		//if (MI->op1_size == 0)
-		//	MI->op1_size = MI->imm_size;
 	}
 }
 
@@ -733,6 +990,10 @@ static void printMemReference(MCInst *MI, unsigned Op, SStream *O)
 	int reg;
 
 	if (MI->csh->detail) {
+#ifndef CAPSTONE_DIET
+		uint8_t access[6];
+#endif
+
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].type = X86_OP_MEM;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].size = MI->x86opsize;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.segment = X86_REG_INVALID;
@@ -740,6 +1001,11 @@ static void printMemReference(MCInst *MI, unsigned Op, SStream *O)
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.index = MCOperand_getReg(IndexReg);
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.scale = (int)ScaleVal;
 		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].mem.disp = 0;
+
+#ifndef CAPSTONE_DIET
+		get_op_access(MI->csh, MCInst_getOpcode(MI), access, &MI->flat_insn->detail->x86.eflags);
+		MI->flat_insn->detail->x86.operands[MI->flat_insn->detail->x86.op_count].access = access[MI->flat_insn->detail->x86.op_count];
+#endif
 	}
 
 	// If this has a segment register, print it.
@@ -774,25 +1040,18 @@ static void printMemReference(MCInst *MI, unsigned Op, SStream *O)
 		if (DispVal) {
 			if (NeedPlus) {
 				if (DispVal < 0) {
-					if (DispVal <  -HEX_THRESHOLD)
-						SStream_concat(O, " - 0x%"PRIx64, -DispVal);
-					else
-						SStream_concat(O, " - %"PRIu64, -DispVal);
+					SStream_concat0(O, " - ");
+					printImm(MI->csh->syntax, O, -DispVal, true);
 				} else {
-					if (DispVal > HEX_THRESHOLD)
-						SStream_concat(O, " + 0x%"PRIx64, DispVal);
-					else
-						SStream_concat(O, " + %"PRIu64, DispVal);
+					SStream_concat0(O, " + ");
+					printImm(MI->csh->syntax, O, DispVal, true);
 				}
 			} else {
 				// memory reference to an immediate address
 				if (DispVal < 0) {
-					SStream_concat(O, "0x%"PRIx64, arch_masks[MI->csh->mode] & DispVal);
+					printImm(MI->csh->syntax, O, arch_masks[MI->csh->mode] & DispVal, true);
 				} else {
-					if (DispVal > HEX_THRESHOLD)
-						SStream_concat(O, "0x%"PRIx64, DispVal);
-					else
-						SStream_concat(O, "%"PRIu64, DispVal);
+					printImm(MI->csh->syntax, O, DispVal, true);
 				}
 			}
 
@@ -812,6 +1071,24 @@ static void printMemReference(MCInst *MI, unsigned Op, SStream *O)
 		MI->op1_size = MI->x86opsize;
 }
 
+static void printanymem(MCInst *MI, unsigned OpNo, SStream *O)
+{
+	switch(MI->Opcode) {
+		default: break;
+		case X86_LEA16r:
+				 MI->x86opsize = 2;
+				 break;
+		case X86_LEA32r:
+		case X86_LEA64_32r:
+				 MI->x86opsize = 4;
+				 break;
+		case X86_LEA64r:
+				 MI->x86opsize = 8;
+				 break;
+	}
+	printMemReference(MI, OpNo, O);
+}
+
 #define GET_REGINFO_ENUM
 #include "X86GenRegisterInfo.inc"
 
