@@ -8,7 +8,7 @@
 //===----------------------------------------------------------------------===//
 
 /* Capstone Disassembly Engine */
-/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2014 */
+/* By Nguyen Anh Quynh <aquynh@gmail.com>, 2013-2015 */
 
 #ifdef CAPSTONE_HAS_XCORE
 
@@ -50,90 +50,91 @@ static bool readInstruction32(const uint8_t *code, size_t code_len, uint32_t *in
 		return false;
 
 	// Encoded as a little-endian 32-bit word in the stream.
-	*insn = (code[0] << 0) | (code[1] << 8) | (code[2] << 16) | (code[3] << 24);
+	*insn = (code[0] << 0) | (code[1] << 8) | (code[2] << 16) | ((uint32_t) code[3] << 24);
+
 	return true;
 }
 
-static unsigned getReg(MCRegisterInfo *MRI, unsigned RC, unsigned RegNo)
+static unsigned getReg(const MCRegisterInfo *MRI, unsigned RC, unsigned RegNo)
 {
-	MCRegisterClass *rc = MCRegisterInfo_getRegClass(MRI, RC);
+	const MCRegisterClass *rc = MCRegisterInfo_getRegClass(MRI, RC);
 	return rc->RegsBegin[RegNo];
 }
 
 static DecodeStatus DecodeGRRegsRegisterClass(MCInst *Inst, unsigned RegNo,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeRRegsRegisterClass(MCInst *Inst, unsigned RegNo,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeBitpOperand(MCInst *Inst, unsigned Val,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeNegImmOperand(MCInst *Inst, unsigned Val,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode2RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode2RImmInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeR2RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode2RSrcDstInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeRUSInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeRUSBitpInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeRUSSrcDstBitpInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL2RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeLR2RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode3RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode3RImmInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode2RUSInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus Decode2RUSBitpInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL3RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL3RSrcDstInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL2RUSInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL2RUSBitpInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL6RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL5RInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL4RSrcDstInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 static DecodeStatus DecodeL4RSrcDstSrcDstInstruction(MCInst *Inst, unsigned Insn,
-		uint64_t Address, void *Decoder);
+		uint64_t Address, const void *Decoder);
 
 #include "XCoreGenDisassemblerTables.inc"
 
@@ -142,7 +143,7 @@ static DecodeStatus DecodeL4RSrcDstSrcDstInstruction(MCInst *Inst, unsigned Insn
 #include "XCoreGenRegisterInfo.inc"
 
 static DecodeStatus DecodeGRRegsRegisterClass(MCInst *Inst, unsigned RegNo,
-		uint64_t Address, void *Decoder)
+		uint64_t Address, const void *Decoder)
 {
 	unsigned Reg;
 
@@ -156,7 +157,7 @@ static DecodeStatus DecodeGRRegsRegisterClass(MCInst *Inst, unsigned RegNo,
 }
 
 static DecodeStatus DecodeRRegsRegisterClass(MCInst *Inst, unsigned RegNo,
-		uint64_t Address, void *Decoder)
+		uint64_t Address, const void *Decoder)
 {
 	unsigned Reg;
 	if (RegNo > 15)
@@ -169,9 +170,9 @@ static DecodeStatus DecodeRRegsRegisterClass(MCInst *Inst, unsigned RegNo,
 }
 
 static DecodeStatus DecodeBitpOperand(MCInst *Inst, unsigned Val,
-		uint64_t Address, void *Decoder)
+		uint64_t Address, const void *Decoder)
 {
-	static unsigned Values[] = {
+	static const unsigned Values[] = {
 		32 /*bpw*/, 1, 2, 3, 4, 5, 6, 7, 8, 16, 24, 32
 	};
 
@@ -183,7 +184,7 @@ static DecodeStatus DecodeBitpOperand(MCInst *Inst, unsigned Val,
 }
 
 static DecodeStatus DecodeNegImmOperand(MCInst *Inst, unsigned Val,
-		uint64_t Address, void *Decoder)
+		uint64_t Address, const void *Decoder)
 {
 	MCOperand_CreateImm0(Inst, -(int64_t)Val);
 	return MCDisassembler_Success;
@@ -233,7 +234,7 @@ static DecodeStatus Decode3OpInstruction(unsigned Insn,
 #define GET_INSTRINFO_ENUM
 #include "XCoreGenInstrInfo.inc"
 static DecodeStatus Decode2OpInstructionFail(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	// Try and decode as a 3R instruction.
 	unsigned Opcode = fieldFromInstruction_4(Insn, 11, 5);
@@ -304,7 +305,7 @@ static DecodeStatus Decode2OpInstructionFail(MCInst *Inst, unsigned Insn, uint64
 }
 
 static DecodeStatus Decode2RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op1, &Op2);
@@ -318,7 +319,7 @@ static DecodeStatus Decode2RInstruction(MCInst *Inst, unsigned Insn, uint64_t Ad
 }
 
 static DecodeStatus Decode2RImmInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op1, &Op2);
@@ -332,7 +333,7 @@ static DecodeStatus Decode2RImmInstruction(MCInst *Inst, unsigned Insn, uint64_t
 }
 
 static DecodeStatus DecodeR2RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op2, &Op1);
@@ -346,7 +347,7 @@ static DecodeStatus DecodeR2RInstruction(MCInst *Inst, unsigned Insn, uint64_t A
 }
 
 static DecodeStatus Decode2RSrcDstInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op1, &Op2);
@@ -361,7 +362,7 @@ static DecodeStatus Decode2RSrcDstInstruction(MCInst *Inst, unsigned Insn, uint6
 }
 
 static DecodeStatus DecodeRUSInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op1, &Op2);
@@ -375,7 +376,7 @@ static DecodeStatus DecodeRUSInstruction(MCInst *Inst, unsigned Insn, uint64_t A
 }
 
 static DecodeStatus DecodeRUSBitpInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op1, &Op2);
@@ -389,7 +390,7 @@ static DecodeStatus DecodeRUSBitpInstruction(MCInst *Inst, unsigned Insn, uint64
 }
 
 static DecodeStatus DecodeRUSSrcDstBitpInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(Insn, &Op1, &Op2);
@@ -404,7 +405,7 @@ static DecodeStatus DecodeRUSSrcDstBitpInstruction(MCInst *Inst, unsigned Insn,
 }
 
 static DecodeStatus DecodeL2OpInstructionFail(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	// Try and decode as a L3R / L2RUS instruction.
 	unsigned Opcode = fieldFromInstruction_4(Insn, 16, 4) |
@@ -476,7 +477,7 @@ static DecodeStatus DecodeL2OpInstructionFail(MCInst *Inst, unsigned Insn, uint6
 }
 
 static DecodeStatus DecodeL2RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(fieldFromInstruction_4(Insn, 0, 16), &Op1, &Op2);
@@ -490,7 +491,7 @@ static DecodeStatus DecodeL2RInstruction(MCInst *Inst, unsigned Insn, uint64_t A
 }
 
 static DecodeStatus DecodeLR2RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2;
 	DecodeStatus S = Decode2OpInstruction(fieldFromInstruction_4(Insn, 0, 16), &Op1, &Op2);
@@ -504,7 +505,7 @@ static DecodeStatus DecodeLR2RInstruction(MCInst *Inst, unsigned Insn, uint64_t
 }
 
 static DecodeStatus Decode3RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S = Decode3OpInstruction(Insn, &Op1, &Op2, &Op3);
@@ -518,7 +519,7 @@ static DecodeStatus Decode3RInstruction(MCInst *Inst, unsigned Insn, uint64_t Ad
 }
 
 static DecodeStatus Decode3RImmInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S = Decode3OpInstruction(Insn, &Op1, &Op2, &Op3);
@@ -532,7 +533,7 @@ static DecodeStatus Decode3RImmInstruction(MCInst *Inst, unsigned Insn, uint64_t
 }
 
 static DecodeStatus Decode2RUSInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S = Decode3OpInstruction(Insn, &Op1, &Op2, &Op3);
@@ -546,7 +547,7 @@ static DecodeStatus Decode2RUSInstruction(MCInst *Inst, unsigned Insn, uint64_t
 }
 
 static DecodeStatus Decode2RUSBitpInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S = Decode3OpInstruction(Insn, &Op1, &Op2, &Op3);
@@ -560,7 +561,7 @@ static DecodeStatus Decode2RUSBitpInstruction(MCInst *Inst, unsigned Insn, uint6
 }
 
 static DecodeStatus DecodeL3RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S =
@@ -575,7 +576,7 @@ static DecodeStatus DecodeL3RInstruction(MCInst *Inst, unsigned Insn, uint64_t A
 }
 
 static DecodeStatus DecodeL3RSrcDstInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S =
@@ -591,7 +592,7 @@ static DecodeStatus DecodeL3RSrcDstInstruction(MCInst *Inst, unsigned Insn, uint
 }
 
 static DecodeStatus DecodeL2RUSInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S =
@@ -606,7 +607,7 @@ static DecodeStatus DecodeL2RUSInstruction(MCInst *Inst, unsigned Insn, uint64_t
 }
 
 static DecodeStatus DecodeL2RUSBitpInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	DecodeStatus S =
@@ -621,7 +622,7 @@ static DecodeStatus DecodeL2RUSBitpInstruction(MCInst *Inst, unsigned Insn, uint
 }
 
 static DecodeStatus DecodeL6RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3, Op4, Op5, Op6;
 	DecodeStatus S =
@@ -643,7 +644,7 @@ static DecodeStatus DecodeL6RInstruction(MCInst *Inst, unsigned Insn, uint64_t A
 }
 
 static DecodeStatus DecodeL5RInstructionFail(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Opcode;
 
@@ -662,7 +663,7 @@ static DecodeStatus DecodeL5RInstructionFail(MCInst *Inst, unsigned Insn, uint64
 }
 
 static DecodeStatus DecodeL5RInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3, Op4, Op5;
 	DecodeStatus S =
@@ -683,7 +684,7 @@ static DecodeStatus DecodeL5RInstruction(MCInst *Inst, unsigned Insn, uint64_t A
 }
 
 static DecodeStatus DecodeL4RSrcDstInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	unsigned Op4 = fieldFromInstruction_4(Insn, 16, 4);
@@ -703,7 +704,7 @@ static DecodeStatus DecodeL4RSrcDstInstruction(MCInst *Inst, unsigned Insn, uint
 }
 
 static DecodeStatus DecodeL4RSrcDstSrcDstInstruction(MCInst *Inst, unsigned Insn, uint64_t Address,
-		void *Decoder)
+		const void *Decoder)
 {
 	unsigned Op1, Op2, Op3;
 	unsigned Op4 = fieldFromInstruction_4(Insn, 16, 4);
@@ -738,7 +739,7 @@ bool XCore_getInstruction(csh ud, const uint8_t *code, size_t code_len, MCInst *
 	}
 
 	if (MI->flat_insn->detail) {
-		memset(MI->flat_insn->detail, 0, sizeof(cs_detail));
+		memset(MI->flat_insn->detail, 0, offsetof(cs_detail, xcore)+sizeof(cs_xcore));
 	}
 
 	// Calling the auto-generated decoder function.
